from enum import Enum
from logging import getLogger
from typing import List, Optional, Union

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3

from gnosis.eth import EthereumClient, EthereumTxSent
from gnosis.eth.contracts import get_multi_send_contract
from gnosis.eth.typing import EthereumData
from gnosis.eth.utils import (
    fast_bytes_to_checksum_address,
    fast_is_checksum_address,
    get_empty_tx_params,
)

logger = getLogger(__name__)


class MultiSendOperation(Enum):
    CALL = 0
    DELEGATE_CALL = 1


class MultiSendTx:
    """
    Wrapper for a single MultiSendTx
    """

    def __init__(
        self,
        operation: MultiSendOperation,
        to: ChecksumAddress,
        value: int,
        data: EthereumData,
        old_encoding: bool = False,
    ):
        """
        :param operation: MultisendOperation, CALL or DELEGATE_CALL
        :param to: Address
        :param value: Value in Wei
        :param data: data as hex string or bytes
        :param old_encoding: True if using old multisend ABI Encoded data, False otherwise
        """
        self.operation = operation
        self.to = to
        self.value = value
        self.data = HexBytes(data) if data else b""
        self.old_encoding = old_encoding

    def __eq__(self, other):
        if not isinstance(other, MultiSendTx):
            return NotImplemented

        return (
            self.operation == other.operation
            and self.to == other.to
            and self.value == other.value
            and self.data == other.data
        )

    def __len__(self):
        """
        :return: Size on bytes of the tx
        """
        return 21 + 32 * 2 + self.data_length

    def __repr__(self):
        data = self.data[:4].hex() + ("..." if len(self.data) > 4 else "")
        return (
            f"MultisendTx operation={self.operation.name} to={self.to} value={self.value} "
            f"data={data}"
        )

    @property
    def data_length(self) -> int:
        return len(self.data)

    @property
    def encoded_data(self):
        operation = HexBytes("{:0>2x}".format(self.operation.value))  # Operation 1 byte
        to = HexBytes("{:0>40x}".format(int(self.to, 16)))  # Address 20 bytes
        value = HexBytes("{:0>64x}".format(self.value))  # Value 32 bytes
        data_length = HexBytes(
            "{:0>64x}".format(self.data_length)
        )  # Data length 32 bytes
        return operation + to + value + data_length + self.data

    @classmethod
    def from_bytes(cls, encoded_multisend_tx: Union[str, bytes]) -> "MultiSendTx":
        """
        Decoded one MultiSend transaction. ABI must be used to get the `transactions` parameter and use that data
        for this function
        :param encoded_multisend_tx:
        :return:
        """
        encoded_multisend_tx = HexBytes(encoded_multisend_tx)
        try:
            return cls._decode_multisend_data(encoded_multisend_tx)
        except ValueError:
            # Try using the old decoding method
            return cls._decode_multisend_old_transaction(encoded_multisend_tx)

    @classmethod
    def _decode_multisend_data(cls, encoded_multisend_tx: Union[str, bytes]):
        """
        Decodes one Multisend transaction. If there's more data after `data` it's ignored. Fallbacks to the old
        multisend structure if this structure cannot be decoded.
        https://etherscan.io/address/0x8D29bE29923b68abfDD21e541b9374737B49cdAD#code
        Structure:
            - operation   -> MultiSendOperation 1 byte
            - to          -> ethereum address 20 bytes
            - value       -> tx value 32 bytes
            - data_length -> 32 bytes
            - data        -> `data_length` bytes
        :param encoded_multisend_tx: 1 multisend transaction encoded
        :return: Tx as a MultisendTx
        """
        encoded_multisend_tx = HexBytes(encoded_multisend_tx)
        operation = MultiSendOperation(encoded_multisend_tx[0])
        to = fast_bytes_to_checksum_address(encoded_multisend_tx[1 : 1 + 20])
        value = int.from_bytes(encoded_multisend_tx[21 : 21 + 32], byteorder="big")
        data_length = int.from_bytes(
            encoded_multisend_tx[21 + 32 : 21 + 32 * 2], byteorder="big"
        )
        data = encoded_multisend_tx[21 + 32 * 2 : 21 + 32 * 2 + data_length]
        len_data = len(data)
        if len_data != data_length:
            raise ValueError(
                f"Data length {data_length} is different from len(data) {len_data}"
            )
        return cls(operation, to, value, data, old_encoding=False)

    @classmethod
    def _decode_multisend_old_transaction(
        cls, encoded_multisend_tx: Union[str, bytes]
    ) -> "MultiSendTx":
        """
        Decodes one old multisend transaction. If there's more data after `data` it's ignored. The difference with
        the new MultiSend is that every value but `data` is padded to 32 bytes, wasting a lot of bytes.
        https://etherscan.io/address/0xE74d6AF1670FB6560dd61EE29eB57C7Bc027Ce4E#code
        Structure:
            - operation   -> MultiSendOperation 32 byte
            - to          -> ethereum address 32 bytes
            - value       -> tx value 32 bytes
            - data_length -> 32 bytes
            - data        -> `data_length` bytes
        :param encoded_multisend_tx: 1 multisend transaction encoded
        :return: Tx as a MultisendTx
        """
        encoded_multisend_tx = HexBytes(encoded_multisend_tx)
        operation = MultiSendOperation(
            int.from_bytes(encoded_multisend_tx[:32], byteorder="big")
        )
        to = fast_bytes_to_checksum_address(encoded_multisend_tx[32:64][-20:])
        value = int.from_bytes(encoded_multisend_tx[64:96], byteorder="big")
        data_length = int.from_bytes(encoded_multisend_tx[128:160], byteorder="big")
        data = encoded_multisend_tx[160 : 160 + data_length]
        len_data = len(data)
        if len_data != data_length:
            raise ValueError(
                f"Data length {data_length} is different from len(data) {len_data}"
            )
        return cls(operation, to, value, data, old_encoding=True)


class MultiSend:
    dummy_w3 = Web3()
    MULTISEND_ADDRESSES = (
        "0xA238CBeb142c10Ef7Ad8442C6D1f9E89e07e7761",  # MultiSend v1.3.0
        "0x998739BFdAAdde7C933B942a68053933098f9EDa",  # MultiSend v1.3.0 (EIP-155)
    )
    MULTISEND_CALL_ONLY_ADDRESSES = (
        "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D",  # MultiSend Call Only v1.3.0
        "0xA1dabEF33b3B82c7814B6D82A79e50F4AC44102B",  # MultiSend Call Only v1.3.0 (EIP-155)
    )

    def __init__(
        self,
        ethereum_client: Optional[EthereumClient] = None,
        address: Optional[ChecksumAddress] = None,
        call_only: bool = True,
    ):
        """
        :param ethereum_client: Required for detecting the address in the network.
        :param address: If not provided, will try to detect it from the hardcoded addresses using `ethereum_client`.
        :param call_only: If `True` use `call only` MultiSend, otherwise use regular one.
            Only if `address` not provided
        """

        self.address = address
        self.ethereum_client = ethereum_client
        self.call_only = call_only
        addresses = (
            self.MULTISEND_CALL_ONLY_ADDRESSES
            if call_only
            else self.MULTISEND_ADDRESSES
        )

        if address:
            assert fast_is_checksum_address(address), (
                "%s proxy factory address not valid" % address
            )
        elif ethereum_client:
            # Try to detect MultiSend address if not provided
            for address in addresses:
                if ethereum_client.is_contract(address):
                    self.address = address
                    break
        else:
            self.address = addresses[0]

        if not self.address:
            raise ValueError(
                f"Cannot find a MultiSend contract for chainId={self.ethereum_client.get_chain_id()}"
            )

    @property
    def w3(self):
        return (self.ethereum_client and self.ethereum_client.w3) or Web3()

    @classmethod
    def from_bytes(cls, encoded_multisend_txs: Union[str, bytes]) -> List[MultiSendTx]:
        """
        Decodes one or more multisend transactions from `bytes transactions` (Abi decoded)

        :param encoded_multisend_txs:
        :return: List of MultiSendTxs
        """
        if not encoded_multisend_txs:
            return []

        encoded_multisend_txs = HexBytes(encoded_multisend_txs)
        multisend_tx = MultiSendTx.from_bytes(encoded_multisend_txs)
        multisend_tx_size = len(multisend_tx)

        assert (
            multisend_tx_size > 0
        ), "Multisend tx cannot be empty"  # This should never happen, just in case
        if multisend_tx.old_encoding:
            next_data_position = (
                (multisend_tx.data_length + 0x1F) // 0x20 * 0x20
            ) + 0xA0
        else:
            next_data_position = multisend_tx_size
        remaining_data = encoded_multisend_txs[next_data_position:]

        return [multisend_tx] + cls.from_bytes(remaining_data)

    @classmethod
    def from_transaction_data(
        cls, multisend_data: Union[str, bytes]
    ) -> List[MultiSendTx]:
        """
        Decodes multisend transactions from transaction data (ABI encoded with selector)

        :return:
        """
        try:
            _, data = get_multi_send_contract(cls.dummy_w3).decode_function_input(
                multisend_data
            )
            return cls.from_bytes(data["transactions"])
        except ValueError:
            return []

    @staticmethod
    def deploy_contract(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy proxy factory contract

        :param ethereum_client:
        :param deployer_account: Ethereum Account
        :return: ``EthereumTxSent`` with the deployed contract address
        """
        contract = get_multi_send_contract(ethereum_client.w3)
        tx = contract.constructor().build_transaction(
            {"from": deployer_account.address}
        )

        tx_hash = ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=120)
        assert tx_receipt
        assert tx_receipt["status"]
        contract_address = tx_receipt["contractAddress"]
        logger.info(
            "Deployed and initialized Proxy Factory Contract=%s by %s",
            contract_address,
            deployer_account.address,
        )
        return EthereumTxSent(tx_hash, tx, contract_address)

    def get_contract(self):
        return get_multi_send_contract(self.w3, self.address)

    def build_tx_data(self, multi_send_txs: List[MultiSendTx]) -> bytes:
        """
        Txs don't need to be valid to get through

        :param multi_send_txs:
        :return:
        """
        multisend_contract = self.get_contract()
        encoded_multisend_data = b"".join([x.encoded_data for x in multi_send_txs])
        return multisend_contract.functions.multiSend(
            encoded_multisend_data
        ).build_transaction(get_empty_tx_params())["data"]



import logging
from typing import List, NamedTuple, Optional, Sequence

from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from hexbytes import HexBytes

from gnosis.eth import EthereumClient, EthereumTxSent
from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.contracts import (
    get_compatibility_fallback_handler_contract,
    get_delegate_constructor_proxy_contract,
    get_safe_contract,
    get_simulate_tx_accessor_V1_4_1_contract,
)
from gnosis.eth.utils import get_empty_tx_params

from .exceptions import InvalidPaymentToken
from .proxy_factory import ProxyFactory
from .safe_create2_tx import InvalidERC20Token, SafeCreate2Tx, SafeCreate2TxBuilder

logger = logging.getLogger(__name__)


class SafeCreationEstimate(NamedTuple):
    gas: int
    gas_price: int
    payment: int
    payment_token: Optional[str]


class SafeCreator:
    @staticmethod
    def create(
        ethereum_client: EthereumClient,
        deployer_account: LocalAccount,
        master_copy_address: ChecksumAddress,
        owners: List[ChecksumAddress],
        threshold: int,
        fallback_handler: Optional[ChecksumAddress] = NULL_ADDRESS,
        proxy_factory_address: Optional[ChecksumAddress] = None,
        payment_token: Optional[ChecksumAddress] = NULL_ADDRESS,
        payment: int = 0,
        payment_receiver: Optional[ChecksumAddress] = NULL_ADDRESS,
    ) -> EthereumTxSent:
        """
        Deploy new Safe proxy pointing to the specified `master_copy` address and configured
        with the provided `owners` and `threshold`. By default, payment for the deployer of the tx will be `0`.
        If `proxy_factory_address` is set deployment will be done using the proxy factory instead of calling
        the `constructor` of a new `DelegatedProxy`
        Using `proxy_factory_address` is recommended

        :param ethereum_client:
        :param deployer_account:
        :param master_copy_address:
        :param owners:
        :param threshold:
        :param fallback_handler:
        :param proxy_factory_address:
        :param payment_token:
        :param payment:
        :param payment_receiver:
        :return:
        """

        assert owners, "At least one owner must be set"
        assert 1 <= threshold <= len(owners), "Threshold=%d must be <= %d" % (
            threshold,
            len(owners),
        )

        initializer = (
            get_safe_contract(ethereum_client.w3, NULL_ADDRESS)
            .functions.setup(
                owners,
                threshold,
                NULL_ADDRESS,  # Contract address for optional delegate call
                b"",  # Data payload for optional delegate call
                fallback_handler,  # Handler for fallback calls to this contract,
                payment_token,
                payment,
                payment_receiver,
            )
            .build_transaction(get_empty_tx_params())["data"]
        )

        if proxy_factory_address:
            proxy_factory = ProxyFactory(proxy_factory_address, ethereum_client)
            return proxy_factory.deploy_proxy_contract_with_nonce(
                deployer_account, master_copy_address, initializer=HexBytes(initializer)
            )

        proxy_contract = get_delegate_constructor_proxy_contract(ethereum_client.w3)
        tx = proxy_contract.constructor(
            master_copy_address, initializer
        ).build_transaction({"from": deployer_account.address})
        tx_hash = ethereum_client.send_unsigned_transaction(
            tx, private_key=deployer_account.key
        )
        tx_receipt = ethereum_client.get_transaction_receipt(tx_hash, timeout=60)
        assert tx_receipt
        assert tx_receipt["status"]

        contract_address = tx_receipt["contractAddress"]
        return EthereumTxSent(tx_hash, tx, contract_address)

    @staticmethod
    def deploy_compatibility_fallback_handler(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy Last compatibility Fallback handler

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: ``EthereumTxSent`` with the deployed contract address
        """

        contract = get_compatibility_fallback_handler_contract(ethereum_client.w3)
        constructor_data = contract.constructor().build_transaction(
            get_empty_tx_params()
        )["data"]
        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data
        )
        logger.info(
            "Deployed and initialized Compatibility Fallback Handler version=%s on address %s by %s",
            "1.4.1",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    @staticmethod
    def deploy_simulate_tx_accessor(
        ethereum_client: EthereumClient, deployer_account: LocalAccount
    ) -> EthereumTxSent:
        """
        Deploy Last compatibility Fallback handler

        :param ethereum_client:
        :param deployer_account: Ethereum account
        :return: ``EthereumTxSent`` with the deployed contract address
        """

        contract = get_simulate_tx_accessor_V1_4_1_contract(ethereum_client.w3)
        constructor_data = contract.constructor().build_transaction(
            get_empty_tx_params()
        )["data"]
        ethereum_tx_sent = ethereum_client.deploy_and_initialize_contract(
            deployer_account, constructor_data
        )
        logger.info(
            "Deployed and initialized Simulate Tx Accessor contract version=%s on address %s by %s",
            "1.4.1",
            ethereum_tx_sent.contract_address,
            deployer_account.address,
        )
        return ethereum_tx_sent

    @staticmethod
    def estimate_safe_creation_2(
        ethereum_client: EthereumClient,
        master_copy_address: ChecksumAddress,
        proxy_factory_address: ChecksumAddress,
        number_owners: int,
        gas_price: int,
        payment_token: Optional[ChecksumAddress],
        payment_receiver: ChecksumAddress = NULL_ADDRESS,
        fallback_handler: Optional[ChecksumAddress] = None,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ) -> SafeCreationEstimate:
        """
        :param ethereum_client:
        :param master_copy_address:
        :param proxy_factory_address:
        :param number_owners:
        :param gas_price:
        :param payment_token:
        :param payment_receiver:
        :param fallback_handler:
        :param payment_token_eth_value:
        :param fixed_creation_cost:
        :return: An estimation for creating a Safe with the provided parameters
        """
        salt_nonce = 15
        owners = [Account.create().address for _ in range(number_owners)]
        threshold = number_owners
        if not fallback_handler:
            fallback_handler = (
                Account.create().address
            )  # Better estimate it, it's required for new Safes
        safe_creation_tx = SafeCreate2TxBuilder(
            w3=ethereum_client.w3,
            master_copy_address=master_copy_address,
            proxy_factory_address=proxy_factory_address,
        ).build(
            owners=owners,
            threshold=threshold,
            fallback_handler=fallback_handler,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
            payment_receiver=payment_receiver,
            payment_token=payment_token,
            payment_token_eth_value=payment_token_eth_value,
            fixed_creation_cost=fixed_creation_cost,
        )
        return SafeCreationEstimate(
            safe_creation_tx.gas,
            safe_creation_tx.gas_price,
            safe_creation_tx.payment,
            safe_creation_tx.payment_token,
        )

    @staticmethod
    def build_safe_create2_tx(
        ethereum_client: EthereumClient,
        master_copy_address: ChecksumAddress,
        proxy_factory_address: ChecksumAddress,
        salt_nonce: int,
        owners: Sequence[ChecksumAddress],
        threshold: int,
        gas_price: int,
        payment_token: Optional[ChecksumAddress],
        payment_receiver: Optional[
            ChecksumAddress
        ] = None,  # If none, it will be `tx.origin`
        fallback_handler: Optional[ChecksumAddress] = NULL_ADDRESS,
        payment_token_eth_value: float = 1.0,
        fixed_creation_cost: Optional[int] = None,
    ) -> SafeCreate2Tx:
        """
        Prepare safe proxy deployment for being relayed. It calculates and sets the costs of deployment to be returned
        to the sender of the tx. If you are an advanced user you may prefer to use `create` function
        """
        try:
            safe_creation_tx = SafeCreate2TxBuilder(
                w3=ethereum_client.w3,
                master_copy_address=master_copy_address,
                proxy_factory_address=proxy_factory_address,
            ).build(
                owners=owners,
                threshold=threshold,
                fallback_handler=fallback_handler,
                salt_nonce=salt_nonce,
                gas_price=gas_price,
                payment_receiver=payment_receiver,
                payment_token=payment_token,
                payment_token_eth_value=payment_token_eth_value,
                fixed_creation_cost=fixed_creation_cost,
            )
        except InvalidERC20Token as exc:
            raise InvalidPaymentToken(
                "Invalid payment token %s" % payment_token
            ) from exc

        return safe_creation_tx



from django.test import TestCase

import pytest
from eth_account import Account

from ... import EthereumClient
from ...oracles import CannotGetPriceFromOracle, UniswapV3Oracle
from ..ethereum_test_case import EthereumTestCaseMixin
from ..test_oracles import (
    dai_token_mainnet_address,
    usdc_token_mainnet_address,
    usdt_token_mainnet_address,
    weth_token_mainnet_address,
)
from ..utils import just_test_if_mainnet_node


class TestUniswapV3Oracle(EthereumTestCaseMixin, TestCase):
    @pytest.mark.xfail(reason="Could fail due to liquidity changes")
    def test_get_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(UniswapV3Oracle.is_available(ethereum_client))

        uniswap_v3_oracle = UniswapV3Oracle(ethereum_client)

        price = uniswap_v3_oracle.get_price(weth_token_mainnet_address)
        self.assertEqual(price, 1.0)

        price = uniswap_v3_oracle.get_price(dai_token_mainnet_address)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        price = uniswap_v3_oracle.get_price(
            weth_token_mainnet_address, dai_token_mainnet_address
        )
        self.assertGreater(price, 1)

        # Test with 2 stablecoins with same decimals
        price = uniswap_v3_oracle.get_price(
            usdt_token_mainnet_address, usdc_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)

        # Test with 2 stablecoins with different decimals
        price = uniswap_v3_oracle.get_price(
            dai_token_mainnet_address, usdc_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)

        price = uniswap_v3_oracle.get_price(
            usdc_token_mainnet_address, dai_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)

        random_token = Account.create().address
        with self.assertRaisesMessage(
            CannotGetPriceFromOracle,
            f"Uniswap V3 pool does not exist for {random_token} and 0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
        ):
            uniswap_v3_oracle.get_price(random_token)

        # Test token with no liquidity
        s_usd_token_mainnet_address = "0x57Ab1ec28D129707052df4dF418D58a2D46d5f51"
        with self.assertRaisesMessage(
            CannotGetPriceFromOracle,
            f"Not enough liquidity on uniswap v3 for pair token_1={s_usd_token_mainnet_address} "
            f"token_2=0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2, at least 2 units of each token are required",
        ):
            uniswap_v3_oracle.get_price(s_usd_token_mainnet_address)

        # Test token with no liquidity
        wrapped_nxm_token_mainnet_address = "0x0d438F3b5175Bebc262bF23753C1E53d03432bDE"
        with self.assertRaisesMessage(
            CannotGetPriceFromOracle,
            f"Not enough liquidity on uniswap v3 for pair token_1={wrapped_nxm_token_mainnet_address} "
            f"token_2=0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2, at least 2 units of each token are required",
        ):
            uniswap_v3_oracle.get_price(wrapped_nxm_token_mainnet_address)

    def test_get_price_contract_not_deployed(self):
        self.assertFalse(UniswapV3Oracle.is_available(self.ethereum_client))
        with self.assertRaisesMessage(
            ValueError,
            f"Uniswap V3 Router Contract {UniswapV3Oracle.DEFAULT_ROUTER_ADDRESS} does not exist",
        ):
            UniswapV3Oracle(self.ethereum_client)



import dataclasses
from typing import Optional

from eth_abi import encode as abi_encode
from eth_abi.packed import encode_packed
from eth_typing import ChecksumAddress
from hexbytes import HexBytes

from gnosis.eth.account_abstraction import UserOperation
from gnosis.eth.utils import fast_keccak

_domain_separator_cache = {}


@dataclasses.dataclass
class SafeOperation:
    """
    Safe EIP4337 operation
    More info: https://github.com/safe-global/safe-modules/blob/main/modules/4337/contracts/Safe4337Module.sol#L55

    TypeHash calculation:

    .. code-block:: python

        keccak256(
            "SafeOp(address safe,uint256 nonce,bytes initCode,bytes callData,uint256 callGasLimit,uint256 verificationGasLimit,uint256 preVerificationGas,uint256 maxFeePerGas,uint256 maxPriorityFeePerGas,bytes paymasterAndData,uint48 validAfter,uint48 validUntil,address entryPoint)"
        )

    Domain separator calculation:

    .. code-block:: python
        keccak256("EIP712Domain(uint256 chainId,address verifyingContract)"

    """

    safe: ChecksumAddress
    nonce: int  # uint256
    init_code_hash: bytes  # bytes32
    call_data_hash: bytes  # bytes32
    call_gas_limit: int  # uint256
    verification_gas_limit: int  # uint256
    pre_verification_gas: int  # uint256
    max_fee_per_gas: int  # uint256
    max_priority_fee_per_gas: int  # uint256
    paymaster_and_data_hash: bytes  # bytes32
    valid_after: int  # uint48
    valid_until: int  # uint48
    entry_point: ChecksumAddress
    signature: bytes
    TYPE_HASH: bytes = HexBytes(
        "0x84aa190356f56b8c87825f54884392a9907c23ee0f8e1ea86336b763faf021bd"
    )  # bytes32
    DOMAIN_SEPARATOR_TYPE_HASH: bytes = HexBytes(
        "0x47e79534a245952e8b16893a336b85a3d9ea9fa8c573f3d803afb92a79469218"
    )  # bytes32
    safe_operation_hash: Optional[bytes] = None
    safe_operation_hash_preimage: Optional[bytes] = None

    @classmethod
    def from_user_operation(cls, user_operation: UserOperation):
        return cls(
            user_operation.sender,
            user_operation.nonce,
            fast_keccak(user_operation.init_code),
            fast_keccak(user_operation.call_data),
            user_operation.call_gas_limit,
            user_operation.verification_gas_limit,
            user_operation.pre_verification_gas,
            user_operation.max_fee_per_gas,
            user_operation.max_priority_fee_per_gas,
            fast_keccak(user_operation.paymaster_and_data),
            int.from_bytes(user_operation.signature[:6], byteorder="big"),
            int.from_bytes(user_operation.signature[6:12], byteorder="big"),
            user_operation.entry_point,
            user_operation.signature[12:],
        )

    def get_domain_separator(
        self, chain_id: int, module_address: ChecksumAddress
    ) -> bytes:
        key = (chain_id, module_address)
        if key not in _domain_separator_cache:
            _domain_separator_cache[key] = fast_keccak(
                abi_encode(
                    ["bytes32", "uint256", "address"],
                    [self.DOMAIN_SEPARATOR_TYPE_HASH, chain_id, module_address],
                )
            )
        return _domain_separator_cache[key]

    def get_safe_operation_hash_preimage(
        self, chain_id: int, module_address: ChecksumAddress
    ) -> bytes:
        # Cache safe_operation_hash
        if not self.safe_operation_hash_preimage:
            encoded_safe_op_struct = abi_encode(
                [
                    "bytes32",
                    "address",
                    "uint256",
                    "bytes32",
                    "bytes32",
                    "uint256",
                    "uint256",
                    "uint256",
                    "uint256",
                    "uint256",
                    "bytes32",
                    "uint48",
                    "uint48",
                    "address",
                ],
                [
                    self.TYPE_HASH,
                    self.safe,
                    self.nonce,
                    self.init_code_hash,
                    self.call_data_hash,
                    self.call_gas_limit,
                    self.verification_gas_limit,
                    self.pre_verification_gas,
                    self.max_fee_per_gas,
                    self.max_priority_fee_per_gas,
                    self.paymaster_and_data_hash,
                    self.valid_after,
                    self.valid_until,
                    self.entry_point,
                ],
            )
            self.safe_operation_hash_preimage = fast_keccak(encoded_safe_op_struct)
        return self.safe_operation_hash_preimage

    def get_safe_operation_hash(
        self, chain_id: int, module_address: ChecksumAddress
    ) -> bytes:
        # Cache safe_operation_hash
        if not self.safe_operation_hash:
            safe_op_struct_hash = self.get_safe_operation_hash_preimage(
                chain_id, module_address
            )
            operation_data = encode_packed(
                ["bytes1", "bytes1", "bytes32", "bytes32"],
                [
                    bytes.fromhex("19"),
                    bytes.fromhex("01"),
                    self.get_domain_separator(chain_id, module_address),
                    safe_op_struct_hash,
                ],
            )
            self.safe_operation_hash = fast_keccak(operation_data)
        return self.safe_operation_hash



import copy
from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

import requests

from ...account_abstraction import (
    BundlerClient,
    BundlerClientConnectionException,
    BundlerClientResponseException,
    UserOperation,
    UserOperationReceipt,
)
from ..mocks.mock_bundler import (
    safe_4337_user_operation_hash_mock,
    supported_entrypoint_mock,
    user_operation_mock,
    user_operation_receipt_mock,
)


class TestBundlerClient(TestCase):
    def setUp(self):
        self.bundler = BundlerClient("https://localhost")

    @mock.patch.object(requests.Session, "post")
    def test_get_user_operation_by_hash(self, mock_session: MagicMock):
        mock_session.return_value.ok = True
        mock_session.return_value.json = MagicMock(
            return_value={"jsonrpc": "2.0", "id": 1, "result": None}
        )
        user_operation_hash = safe_4337_user_operation_hash_mock.hex()

        self.assertIsNone(self.bundler.get_user_operation_by_hash(user_operation_hash))
        mock_session.return_value.json = MagicMock(return_value=user_operation_mock)
        self.bundler.get_user_operation_by_hash.cache_clear()
        expected_user_operation = UserOperation.from_bundler_response(
            user_operation_hash, user_operation_mock["result"]
        )
        self.assertEqual(
            self.bundler.get_user_operation_by_hash(user_operation_hash),
            expected_user_operation,
        )

        mock_session.return_value.json = MagicMock(
            return_value={
                "jsonrpc": "2.0",
                "id": 1,
                "error": {
                    "code": -32600,
                    "message": "Unspecified origin not on whitelist.",
                },
            }
        )
        self.bundler.get_user_operation_by_hash.cache_clear()
        with self.assertRaises(BundlerClientResponseException):
            self.assertIsNone(
                self.bundler.get_user_operation_by_hash(user_operation_hash)
            )
        mock_session.side_effect = IOError
        with self.assertRaises(BundlerClientConnectionException):
            self.assertIsNone(
                self.bundler.get_user_operation_by_hash(user_operation_hash)
            )

    @mock.patch.object(requests.Session, "post")
    def test_get_user_operation_receipt(self, mock_session: MagicMock):
        mock_session.return_value.ok = True
        mock_session.return_value.json = MagicMock(
            return_value={"jsonrpc": "2.0", "id": 1, "result": None}
        )
        user_operation_hash = safe_4337_user_operation_hash_mock.hex()
        self.assertIsNone(self.bundler.get_user_operation_receipt(user_operation_hash))
        mock_session.return_value.json = MagicMock(
            return_value=copy.deepcopy(user_operation_receipt_mock)
        )
        self.bundler.get_user_operation_receipt.cache_clear()

        expected_user_operation_receipt = UserOperationReceipt.from_bundler_response(
            user_operation_receipt_mock["result"]
        )
        self.assertEqual(
            self.bundler.get_user_operation_receipt(user_operation_hash),
            expected_user_operation_receipt,
        )
        mock_session.return_value.json = MagicMock(
            return_value={
                "jsonrpc": "2.0",
                "id": 1,
                "error": {
                    "code": -32600,
                    "message": "Unspecified origin not on whitelist.",
                },
            }
        )
        self.bundler.get_user_operation_receipt.cache_clear()
        with self.assertRaises(BundlerClientResponseException):
            self.assertIsNone(
                self.bundler.get_user_operation_receipt(user_operation_hash)
            )
        mock_session.side_effect = IOError
        with self.assertRaises(BundlerClientConnectionException):
            self.assertIsNone(
                self.bundler.get_user_operation_receipt(user_operation_hash)
            )

    @mock.patch.object(requests.Session, "post")
    def test_get_user_operation_and_receipt(self, mock_session: MagicMock):
        mock_session.return_value.ok = True
        mock_session.return_value.json = MagicMock(
            return_value=[
                {"jsonrpc": "2.0", "id": 1, "result": None},
                {"jsonrpc": "2.0", "id": 2, "result": None},
            ]
        )
        user_operation_hash = safe_4337_user_operation_hash_mock.hex()
        self.assertIsNone(
            self.bundler.get_user_operation_and_receipt(user_operation_hash)
        )
        mock_session.return_value.json = MagicMock(
            return_value=[
                user_operation_mock,
                copy.deepcopy(user_operation_receipt_mock),
            ]
        )
        self.bundler.get_user_operation_and_receipt.cache_clear()
        expected_user_operation = UserOperation.from_bundler_response(
            user_operation_hash, user_operation_mock["result"]
        )
        expected_user_operation_receipt = UserOperationReceipt.from_bundler_response(
            user_operation_receipt_mock["result"]
        )
        (
            user_operation,
            user_operation_receipt,
        ) = self.bundler.get_user_operation_and_receipt(user_operation_hash)

        self.assertEqual(
            user_operation,
            expected_user_operation,
        )
        self.assertEqual(user_operation_receipt, expected_user_operation_receipt)
        mock_session.return_value.json = MagicMock(
            return_value={
                "jsonrpc": "2.0",
                "id": 1,
                "error": {
                    "code": -32600,
                    "message": "Unspecified origin not on whitelist.",
                },
            }
        )
        self.bundler.get_user_operation_and_receipt.cache_clear()
        with self.assertRaises(BundlerClientResponseException):
            self.assertIsNone(
                self.bundler.get_user_operation_and_receipt(user_operation_hash)
            )
        mock_session.side_effect = IOError
        with self.assertRaises(BundlerClientConnectionException):
            self.assertIsNone(
                self.bundler.get_user_operation_and_receipt(user_operation_hash)
            )

    @mock.patch.object(requests.Session, "post")
    def test_supported_entry_points(self, mock_session: MagicMock):
        mock_session.return_value.ok = True
        mock_session.return_value.json = MagicMock(
            return_value=supported_entrypoint_mock
        )

        self.assertEqual(
            self.bundler.supported_entry_points(), supported_entrypoint_mock["result"]
        )



from typing import List, Tuple, Union

from eth_keys import keys
from eth_keys.exceptions import BadSignature
from hexbytes import HexBytes

from gnosis.eth.constants import NULL_ADDRESS


def signature_split(
    signatures: Union[bytes, str], pos: int = 0
) -> Tuple[int, int, int]:
    """
    :param signatures: signatures in form of {bytes32 r}{bytes32 s}{uint8 v}
    :param pos: position of the signature
    :return: Tuple with v, r, s
    """
    signatures = HexBytes(signatures)
    signature_pos = 65 * pos
    if len(signatures[signature_pos : signature_pos + 65]) < 65:
        raise ValueError(f"Signature must be at least 65 bytes {signatures.hex()}")
    r = int.from_bytes(signatures[signature_pos : 32 + signature_pos], "big")
    s = int.from_bytes(signatures[32 + signature_pos : 64 + signature_pos], "big")
    v = signatures[64 + signature_pos]

    return v, r, s


def signature_to_bytes(v: int, r: int, s: int) -> bytes:
    """
    Convert ecdsa signature to bytes
    :param v:
    :param r:
    :param s:
    :return: signature in form of {bytes32 r}{bytes32 s}{uint8 v}
    """

    return (
        r.to_bytes(32, byteorder="big")
        + s.to_bytes(32, byteorder="big")
        + v.to_bytes(1, byteorder="big")
    )


def signatures_to_bytes(signatures: List[Tuple[int, int, int]]) -> bytes:
    """
    Convert signatures to bytes
    :param signatures: list of tuples(v, r, s)
    :return: 65 bytes per signature
    """
    return b"".join([signature_to_bytes(v, r, s) for v, r, s in signatures])


def get_signing_address(signed_hash: Union[bytes, str], v: int, r: int, s: int) -> str:
    """
    :return: checksummed ethereum address, for example `0x568c93675A8dEb121700A6FAdDdfE7DFAb66Ae4A`
    :rtype: str or `NULL_ADDRESS` if signature is not valid
    """
    try:
        public_key = keys.ecdsa_recover(signed_hash, keys.Signature(vrs=(v - 27, r, s)))
        return public_key.to_checksum_address()
    except BadSignature:
        return NULL_ADDRESS



class BundlerClientException(ValueError):
    pass


class BundlerClientConnectionException(BundlerClientException, IOError):
    pass


class BundlerClientResponseException(BundlerClientException):
    pass



from abc import ABCMeta, abstractmethod
from functools import cached_property
from logging import getLogger
from typing import Callable, Optional

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract

logger = getLogger(__name__)


class ContractBase(metaclass=ABCMeta):
    def __init__(
        self,
        address: ChecksumAddress,
        ethereum_client: "EthereumClient",  # noqa F821
        *args,
        **kwargs
    ):
        self.address = address
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3

    @abstractmethod
    def get_contract_fn(self) -> Callable[[Web3, Optional[ChecksumAddress]], Contract]:
        """
        :return: Contract function to get the proper contract
        """
        raise NotImplementedError

    @cached_property
    def contract(self) -> Contract:
        return self.get_contract_fn()(self.ethereum_client.w3, self.address)



import logging

from hexbytes import HexBytes

from .. import EthereumNetwork
from .oracles import UniswapV2Oracle

logger = logging.getLogger(__name__)


class SushiswapOracle(UniswapV2Oracle):
    PAIR_INIT_CODE = HexBytes(
        "0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303"
    )
    ROUTER_ADDRESSES = {
        EthereumNetwork.MAINNET: "0xd9e1cE17f2641f24aE83637ab66a2cca9C378B9F",
        EthereumNetwork.POLYGON: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.ARBITRUM_ONE: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.AVALANCHE_C_CHAIN: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.MOONRIVER: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.FANTOM_OPERA: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.BNB_SMART_CHAIN_MAINNET: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.GNOSIS: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.CELO_MAINNET: "0x1421bDe4B10e8dd459b3BCb598810B1337D56842",
        EthereumNetwork.FUSE_MAINNET: "0xF4d73326C13a4Fc5FD7A064217e12780e9Bd62c3",
        EthereumNetwork.OKXCHAIN_MAINNET: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.PALM: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.MOONBEAM: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
        EthereumNetwork.HUOBI_ECO_CHAIN_MAINNET: "0x1b02dA8Cb0d097eB8D57A175b88c7D8b47997506",
    }



import logging
from typing import Optional

from gnosis.cowsap import CowSwapAPI, OrderKind

from .. import EthereumClient, EthereumNetworkNotSupported
from .exceptions import CannotGetPriceFromOracle
from .oracles import PriceOracle
from .utils import get_decimals

logger = logging.getLogger(__name__)


class CowswapOracle(PriceOracle):
    """
    CowSwap Oracle implementation

    https://docs.cow.fi/off-chain-services/api
    """

    def __init__(
        self,
        ethereum_client: EthereumClient,
    ):
        """
        :param ethereum_client:
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.api = CowSwapAPI(ethereum_client.get_network())

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if CowSwap is available for the EthereumClient provided, `False` otherwise
        """
        try:
            CowSwapAPI(ethereum_client.get_network())
            return True
        except EthereumNetworkNotSupported:
            return False

    def get_price(
        self, token_address_1: str, token_address_2: Optional[str] = None
    ) -> float:
        token_address_2 = token_address_2 or self.api.weth_address
        if token_address_1 == token_address_2:
            return 1.0

        token_1_decimals = get_decimals(token_address_1, self.ethereum_client)
        try:
            result = self.api.get_estimated_amount(
                token_address_1, token_address_2, OrderKind.SELL, 10**token_1_decimals
            )
            if "buyAmount" in result:
                # Decimals needs to be adjusted
                token_2_decimals = get_decimals(token_address_2, self.ethereum_client)
                return (
                    float(result["buyAmount"])
                    / result["sellAmount"]
                    * 10 ** (token_1_decimals - token_2_decimals)
                )

            exception = None
        except IOError as exc:
            exception = exc
            result = {}

        message = (
            f"Cannot get price from CowSwap {result} "
            f"for token-1={token_address_1} to token-2={token_address_2}"
        )
        logger.debug(message)
        raise CannotGetPriceFromOracle(message) from exception



from django.test import TestCase

import pytest

from ... import EthereumNetwork
from ...clients import BlockscoutClient, BlockScoutConfigurationProblem
from .mocks import sourcify_safe_metadata


class TestBlockscoutClient(TestCase):
    @pytest.mark.flaky(reruns=5)
    def test_blockscout_client(self):
        with self.assertRaises(BlockScoutConfigurationProblem):
            BlockscoutClient(EthereumNetwork.MAINNET)

        blockscout_client = BlockscoutClient(EthereumNetwork.GNOSIS)
        safe_master_copy_abi = sourcify_safe_metadata["output"]["abi"]
        safe_master_copy_address = "0x6851D6fDFAfD08c0295C392436245E5bc78B0185"
        contract_metadata = blockscout_client.get_contract_metadata(
            safe_master_copy_address
        )
        self.assertEqual(contract_metadata.name, "GnosisSafe")
        self.assertEqual(contract_metadata.abi, safe_master_copy_abi)
        random_address = "0xaE32496491b53841efb51829d6f886387708F99a"
        self.assertIsNone(blockscout_client.get_contract_metadata(random_address))



# flake8: noqa F401
from .blockscout_client import (
    BlockscoutClient,
    BlockscoutClientException,
    BlockScoutConfigurationProblem,
)
from .contract_metadata import ContractMetadata
from .ens_client import EnsClient
from .etherscan_client import (
    EtherscanClient,
    EtherscanClientConfigurationProblem,
    EtherscanClientException,
    EtherscanRateLimitError,
)
from .sourcify_client import (
    SourcifyClient,
    SourcifyClientConfigurationProblem,
    SourcifyClientException,
)

__all__ = [
    "BlockScoutConfigurationProblem",
    "BlockscoutClient",
    "BlockscoutClientException",
    "ContractMetadata",
    "EnsClient",
    "EtherscanClient",
    "EtherscanClientConfigurationProblem",
    "EtherscanClientException",
    "EtherscanRateLimitError",
    "SourcifyClient",
    "SourcifyClientConfigurationProblem",
    "SourcifyClientException",
]



AAVE_ATOKEN_ABI = [
    {
        "inputs": [
            {
                "internalType": "contract ILendingPool",
                "name": "pool",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "underlyingAssetAddress",
                "type": "address",
            },
            {
                "internalType": "address",
                "name": "reserveTreasuryAddress",
                "type": "address",
            },
            {"internalType": "string", "name": "tokenName", "type": "string"},
            {"internalType": "string", "name": "tokenSymbol", "type": "string"},
            {
                "internalType": "address",
                "name": "incentivesController",
                "type": "address",
            },
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "index",
                "type": "uint256",
            },
        ],
        "name": "BalanceTransfer",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "target",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "index",
                "type": "uint256",
            },
        ],
        "name": "Burn",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "underlyingAsset",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "pool",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "treasury",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "incentivesController",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint8",
                "name": "aTokenDecimals",
                "type": "uint8",
            },
            {
                "indexed": False,
                "internalType": "string",
                "name": "aTokenName",
                "type": "string",
            },
            {
                "indexed": False,
                "internalType": "string",
                "name": "aTokenSymbol",
                "type": "string",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "params",
                "type": "bytes",
            },
        ],
        "name": "Initialized",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "index",
                "type": "uint256",
            },
        ],
        "name": "Mint",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "inputs": [],
        "name": "ATOKEN_REVISION",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "DOMAIN_SEPARATOR",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "EIP712_REVISION",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "PERMIT_TYPEHASH",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "POOL",
        "outputs": [
            {"internalType": "contract ILendingPool", "name": "", "type": "address"}
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "RESERVE_TREASURY_ADDRESS",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "UINT_MAX_VALUE",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "UNDERLYING_ASSET_ADDRESS",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "", "type": "address"}],
        "name": "_nonces",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "user", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "user", "type": "address"},
            {
                "internalType": "address",
                "name": "receiverOfUnderlying",
                "type": "address",
            },
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
            {"internalType": "uint256", "name": "index", "type": "uint256"},
        ],
        "name": "burn",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "subtractedValue", "type": "uint256"},
        ],
        "name": "decreaseAllowance",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getIncentivesController",
        "outputs": [
            {
                "internalType": "contract IAaveIncentivesController",
                "name": "",
                "type": "address",
            }
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "user", "type": "address"}],
        "name": "getScaledUserBalanceAndSupply",
        "outputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "addedValue", "type": "uint256"},
        ],
        "name": "increaseAllowance",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "uint8",
                "name": "underlyingAssetDecimals",
                "type": "uint8",
            },
            {"internalType": "string", "name": "tokenName", "type": "string"},
            {"internalType": "string", "name": "tokenSymbol", "type": "string"},
        ],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "user", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
            {"internalType": "uint256", "name": "index", "type": "uint256"},
        ],
        "name": "mint",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
            {"internalType": "uint256", "name": "index", "type": "uint256"},
        ],
        "name": "mintToTreasury",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "permit",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "user", "type": "address"}],
        "name": "scaledBalanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "scaledTotalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "sender", "type": "address"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "transferOnLiquidation",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "target", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "transferUnderlyingTo",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]



import functools
import json
import os
from copy import deepcopy
from typing import Any

import pytest
import requests
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import Contract
from web3.types import TxParams

from ..contracts import get_example_erc20_contract


def just_test_if_node_available(node_url_variable_name: str) -> str:
    """
    Just run the test if ``node url`` is defined on the ``node_url_variable_name`` environment variable
    and it's accessible. Node JSON RPC url will only be tested the first call to this function.

    :param node_url_variable_name: Environment variable name for ``node url``
    :return: ``node url``
    """
    node_url = getattr(just_test_if_node_available, node_url_variable_name, None)
    if node_url:  # Just check node first time
        return node_url

    node_url = os.environ.get(node_url_variable_name)
    if not node_url:
        pytest.skip(
            f"{node_url_variable_name} not defined, skipping test",
            allow_module_level=True,
        )
    else:
        try:
            response = requests.post(
                node_url,
                timeout=5,
                json={
                    "jsonrpc": "2.0",
                    "method": "eth_blockNumber",
                    "params": [],
                    "id": 1,
                },
            )
            if not response.ok:
                pytest.fail(
                    f"Problem connecting to node {node_url}: {response.status_code} - {response.content}"
                )
        except IOError:
            pytest.fail(f"Problem connecting to {node_url}")
    setattr(just_test_if_node_available, node_url_variable_name, node_url)
    return node_url


def just_test_if_mainnet_node() -> str:
    return just_test_if_node_available("ETHEREUM_MAINNET_NODE")


def just_test_if_polygon_node() -> str:
    return just_test_if_node_available("ETHEREUM_POLYGON_NODE")


def skip_on(exception, reason="Test skipped due to a controlled exception"):
    """
    Decorator to skip a test if an exception is raised instead of failing it

    :param exception:
    :param reason:
    :return:
    """

    def decorator_func(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                # Run the test
                return f(*args, **kwargs)
            except exception:
                pytest.skip(reason)

        return wrapper

    return decorator_func


# TODO Move this to EthereumClient
def send_tx(w3: Web3, tx: TxParams, account: LocalAccount) -> bytes:
    tx["from"] = account.address
    if "nonce" not in tx:
        tx["nonce"] = w3.eth.get_transaction_count(
            account.address, block_identifier="pending"
        )

    if "gasPrice" not in tx and "maxFeePerGas" not in tx:
        tx["gasPrice"] = w3.eth.gas_price

    if "gas" not in tx:
        tx["gas"] = w3.eth.estimate_gas(tx)

    signed_tx = account.sign_transaction(tx)
    tx_hash = w3.eth.send_raw_transaction(bytes(signed_tx.rawTransaction))
    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    assert tx_receipt["status"] == 1, "Error with tx %s - %s" % (tx_hash.hex(), tx)
    return tx_hash


def deploy_erc20(
    w3: Web3,
    account: LocalAccount,
    name: str,
    symbol: str,
    owner: ChecksumAddress,
    amount: int,
    decimals: int = 18,
) -> Contract:
    erc20_contract = get_example_erc20_contract(w3)
    tx = erc20_contract.constructor(
        name, symbol, decimals, owner, amount
    ).build_transaction(
        {
            "nonce": w3.eth.get_transaction_count(
                account.address, block_identifier="pending"
            )
        }
    )
    signed_tx = account.sign_transaction(tx)
    tx_hash = w3.eth.send_raw_transaction(signed_tx.rawTransaction)

    tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
    erc20_address = tx_receipt["contractAddress"]
    deployed_erc20 = get_example_erc20_contract(w3, erc20_address)
    assert deployed_erc20.functions.balanceOf(owner).call() == amount
    return deployed_erc20


def bytes_to_str(o: Any) -> Any:
    """
    Converts bytes (and hexbytes) fields to `str` in nested data types

    :param o:
    :return:
    """
    if isinstance(o, bytes):
        return HexBytes(o).hex()
    if isinstance(o, dict):
        o = dict(o)  # Remove AttributeDict
        for k in o.keys():
            o[k] = bytes_to_str(o[k])
    elif isinstance(o, (list, tuple)):
        o = deepcopy(o)
        for i, v in enumerate(o):
            o[i] = bytes_to_str(o[i])
    elif isinstance(o, set):
        o = {bytes_to_str(element) for element in o}
    return o


def to_json_with_hexbytes(o: Any) -> str:
    """
    Convert RPC calls with nested bytes/Hexbytes to json and compare. Useful for RPC calls

    :param o:
    :return: Object as JSON with Hexbytes/bytes parsed correctly as a hex string
    """
    return json.dumps(bytes_to_str(o), indent=4, sort_keys=True)



import logging

from django.test import TestCase

import eth_abi
from eth_abi import encode as encode_abi
from eth_abi.packed import encode_packed
from eth_account import Account
from eth_account.messages import defunct_hash_message
from hexbytes import HexBytes
from web3 import Web3

from gnosis.eth.utils import fast_keccak, fast_keccak_text

from ...eth.tests.ethereum_test_case import EthereumTestCaseMixin
from ..safe_signature import (
    SafeSignature,
    SafeSignatureApprovedHash,
    SafeSignatureContract,
    SafeSignatureEOA,
    SafeSignatureEthSign,
    SafeSignatureType,
)
from ..signatures import signature_to_bytes
from .safe_test_case import SafeTestCaseMixin

logger = logging.getLogger(__name__)


class TestSafeSignature(EthereumTestCaseMixin, TestCase):
    def test_contract_signature(self):
        owner = "0x05c85Ab5B09Eb8A55020d72daf6091E04e264af9"
        safe_tx_hash = HexBytes(
            "0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196"
        )
        signature = HexBytes(
            "0x00000000000000000000000005c85ab5b09eb8a55020d72daf6091e04e264af900000000000000000000000"
            "0000000000000000000000000000000000000000000"
        )
        safe_signature = SafeSignatureContract(signature, safe_tx_hash, b"", b"")
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(
            safe_signature.signature_type, SafeSignatureType.CONTRACT_SIGNATURE
        )
        self.assertTrue(str(safe_signature))  # Test __str__
        self.assertFalse(safe_signature.is_valid(self.ethereum_client))

    def test_approved_hash_signature(self):
        owner = "0x05c85Ab5B09Eb8A55020d72daf6091E04e264af9"
        safe_tx_hash = HexBytes(
            "0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196"
        )
        signature = HexBytes(
            "0x00000000000000000000000005c85ab5b09eb8a55020d72daf6091e04e264af900000000000000000000000"
            "0000000000000000000000000000000000000000001"
        )
        safe_signature = SafeSignatureApprovedHash(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertTrue(str(safe_signature))  # Test __str__

        # Problematic signature found on rinkeby tx-hash 0x10df7fd8b75297af360ae335ca7562be92cafbbd39cc72172e17fcaebccc4f42
        # Note the ff preceding the address
        signature = HexBytes(
            "0x0000000000000000000000ffdc5299b629ef24fdecfbb240c00fc79fabb9cf97000000000000000000000000000000000000000000000000000000000000000001"
        )
        # Safe tx hash is not relevant for this case, use the same one
        safe_signature = SafeSignatureApprovedHash(signature, b"")
        self.assertEqual(
            safe_signature.owner, "0xdC5299b629Ef24fDECfBb240C00Fc79FAbB9cf97"
        )

    def test_approved_hash_signature_build(self):
        owner = Account.create().address
        safe_tx_hash = fast_keccak_text("random")
        safe_signature = SafeSignatureApprovedHash.build_for_owner(owner, safe_tx_hash)
        self.assertEqual(len(safe_signature.signature), 65)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertEqual(safe_signature.owner, owner)

        # Check parse signature get the same result
        safe_signature = SafeSignature.parse_signature(
            safe_signature.signature, safe_tx_hash
        )[0]
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertEqual(safe_signature.owner, owner)

    def test_eth_sign_signature(self):
        account = Account.create()
        owner = account.address
        safe_tx_hash = HexBytes(
            "0x123477d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf195"
        )
        message = defunct_hash_message(primitive=safe_tx_hash)
        signature = account.signHash(message)["signature"]
        signature = signature[:64] + HexBytes(signature[64] + 4)  # Add 4 to v
        safe_signature = SafeSignatureEthSign(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.ETH_SIGN)
        self.assertTrue(safe_signature.is_valid())
        self.assertTrue(str(safe_signature))  # Test __str__

    def test_eoa_signature(self):
        owner = "0xADb7CB706e9A1bd9F96a397da340bF34a9984E1E"
        safe_tx_hash = HexBytes(
            "0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196"
        )
        signature = HexBytes(
            "0x5dccf9f1375cee56331a67867a0b05a828894c2b76dee84546ec663997d7548257cb2d606087ee7590b966e"
            "958d97a65528ae941a0f7e5050949f618629509c81b"
        )
        safe_signature = SafeSignatureEOA(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.EOA)
        self.assertTrue(safe_signature.is_valid())

        account = Account.create()
        owner = account.address
        safe_tx_hash = HexBytes(
            "0x123477d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf195"
        )
        signature = account.signHash(safe_tx_hash)["signature"]
        safe_signature = SafeSignatureEOA(signature, safe_tx_hash)
        self.assertEqual(safe_signature.owner, owner)
        self.assertEqual(safe_signature.signature_type, SafeSignatureType.EOA)
        self.assertTrue(safe_signature.is_valid())
        self.assertTrue(str(safe_signature))  # Test __str__

    def test_defunct_hash_message(self):
        safe_tx_hash = HexBytes(
            "0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196"
        )
        ethereum_signed_message = "\x19Ethereum Signed Message:\n32"
        encoded_message = encode_packed(
            ["(string,bytes32)"], [(ethereum_signed_message, HexBytes(safe_tx_hash))]
        )
        encoded_hash = fast_keccak(encoded_message)
        self.assertEqual(encoded_hash, defunct_hash_message(primitive=safe_tx_hash))

    def test_parse_signature(self):
        owner_1 = "0x05c85Ab5B09Eb8A55020d72daf6091E04e264af9"
        owner_2 = "0xADb7CB706e9A1bd9F96a397da340bF34a9984E1E"
        safe_tx_hash = HexBytes(
            "0x4c9577d1b1b8dec52329a983ae26238b65f74b7dd9fb28d74ad9548e92aaf196"
        )
        signatures = HexBytes(
            "0x00000000000000000000000005c85ab5b09eb8a55020d72daf6091e04e264af90000000000000000000000"
            "000000000000000000000000000000000000000000015dccf9f1375cee56331a67867a0b05a828894c2b76de"
            "e84546ec663997d7548257cb2d606087ee7590b966e958d97a65528ae941a0f7e5050949f618629509c81b"
        )

        s1, s2 = SafeSignature.parse_signature(signatures, safe_tx_hash)
        self.assertEqual(s1.owner, owner_1)
        self.assertEqual(s1.signature_type, SafeSignatureType.APPROVED_HASH)
        self.assertIsInstance(s1, SafeSignatureApprovedHash)
        self.assertEqual(s2.signature_type, SafeSignatureType.EOA)
        self.assertEqual(s2.owner, owner_2)
        self.assertTrue(s2.is_valid())
        self.assertIsInstance(s2, SafeSignatureEOA)

    def test_parse_signature_with_trailing_zeroes(self):
        signatures = HexBytes(
            "0x9173a0b0895c28cdc82729f0a66e3645820a2e395cadad588463ab483e42b3df6e8b05d258e58c5255136"
            "0e032733b507f731001cce6b5391b7750439c8cffc020000000000000000000000000dc5299b629ef24fdec"
            "fbb240c00fc79fabb9cf9700000000000000000000000000000000000000000000000000000000000000000"
            "1000000000000000000000000000000000000000000000000000000000000"
        )
        parsed_signatures = SafeSignature.parse_signature(signatures, "")
        self.assertEqual(len(parsed_signatures), 2)
        self.assertEqual(
            parsed_signatures[0].signature_type, SafeSignatureType.ETH_SIGN
        )
        self.assertEqual(
            parsed_signatures[1].signature_type, SafeSignatureType.APPROVED_HASH
        )

    def test_parse_signature_empty(self):
        safe_tx_hash = fast_keccak_text("Legoshi")
        for value in (b"", "", None):
            self.assertEqual(SafeSignature.parse_signature(value, safe_tx_hash), [])


class TestSafeContractSignature(SafeTestCaseMixin, TestCase):
    def test_contract_signature_for_message(self):
        account = Account.create()
        safe_owner = self.deploy_test_safe(owners=[account.address])
        safe = self.deploy_test_safe(owners=[safe_owner.address])

        safe_address = safe.address
        message = "Testing EIP191 message signing"
        message_hash = defunct_hash_message(text=message)
        safe_owner_message_hash = safe_owner.get_message_hash(message_hash)
        safe_owner_signature = account.signHash(safe_owner_message_hash)["signature"]
        safe_parent_message_hash = safe.get_message_hash(message_hash)

        # Build EIP1271 signature v=0 r=safe v=dynamic_part dynamic_part=size+owner_signature
        signature_1271 = (
            signature_to_bytes(
                0, int.from_bytes(HexBytes(safe_owner.address), byteorder="big"), 65
            )
            + eth_abi.encode(["bytes"], [safe_owner_signature])[32:]
        )

        safe_signatures = SafeSignature.parse_signature(
            signature_1271, safe_parent_message_hash, message_hash
        )
        self.assertEqual(len(safe_signatures), 1)
        self.assertTrue(safe_signatures[0].is_valid(self.ethereum_client))

    def test_export_signatures(self):
        """
        Create a Safe with 3 signers and threshold 3, 1 EOA and 2 Safes.
        1 Safe will have threshold 2 and the other threshold 1, to test different size dynamic parts

        :return:
        """
        #
        safe_owner_1_eoa_1 = Account.create()
        safe_owner_1_eoa_2 = Account.create()
        safe_owner_1 = self.deploy_test_safe(
            owners=[safe_owner_1_eoa_1.address, safe_owner_1_eoa_2.address], threshold=2
        )
        safe_owner_2_eoa_1 = Account.create()
        safe_owner_2 = self.deploy_test_safe(owners=[safe_owner_2_eoa_1.address])
        eoa = Account.create()
        safe = self.deploy_test_safe(
            owners=[safe_owner_1.address, safe_owner_2.address, eoa.address],
            threshold=3,
        )

        to = eoa.address  # Not relevant
        value = 0
        data = HexBytes("")
        safe_tx = safe.build_multisig_tx(
            to=to,
            value=value,
            data=data,
        )
        safe_tx_hash_preimage = safe_tx.safe_tx_hash_preimage
        safe_tx_hash = safe_tx.safe_tx_hash

        safe_owner_1_message_hash = safe_owner_1.get_message_hash(safe_tx_hash_preimage)
        safe_owner_1_eoa_1_signature = safe_owner_1_eoa_1.signHash(
            safe_owner_1_message_hash
        )["signature"]
        safe_owner_1_eoa_2_signature = safe_owner_1_eoa_2.signHash(
            safe_owner_1_message_hash
        )["signature"]
        safe_owner_1_eoa_signature = (
            safe_owner_1_eoa_1_signature + safe_owner_1_eoa_2_signature
            if safe_owner_1_eoa_1.address.lower() < safe_owner_1_eoa_2.address.lower()
            else safe_owner_1_eoa_2_signature + safe_owner_1_eoa_1_signature
        )

        # Build EIP1271 signature v=0 r=safe v=dynamic_part dynamic_part=size+owner_signature
        safe_signature_contract_1 = SafeSignatureContract.from_values(
            safe_owner_1.address,
            safe_tx_hash,
            safe_tx_hash_preimage,
            safe_owner_1_eoa_signature,
        )

        safe_owner_2_message_hash = safe_owner_2.get_message_hash(safe_tx_hash_preimage)
        safe_owner_2_eoa_1_signature = safe_owner_2_eoa_1.signHash(
            safe_owner_2_message_hash
        )["signature"]

        # Build EIP1271 signature v=0 r=safe v=dynamic_part dynamic_part=size+owner_signature
        safe_signature_contract_2 = SafeSignatureContract.from_values(
            safe_owner_2.address,
            safe_tx_hash,
            safe_tx_hash_preimage,
            safe_owner_2_eoa_1_signature,
        )

        eoa_signature = SafeSignatureEOA(
            eoa.signHash(safe_tx_hash)["signature"], safe_tx_hash
        )

        signatures = SafeSignature.export_signatures(
            [safe_signature_contract_1, eoa_signature, safe_signature_contract_2]
        )
        safe_tx.signatures = signatures
        self.assertEqual(safe_tx.call(), 1)
        safe_tx.execute(self.ethereum_test_account.key)

    def test_contract_signature(self):
        owner_1 = self.ethereum_test_account
        safe = self.deploy_test_safe_v1_1_1(
            owners=[owner_1.address], initial_funding_wei=Web3.to_wei(0.01, "ether")
        )
        safe_contract = safe.contract
        safe_tx_hash = fast_keccak_text("test")
        signature_r = HexBytes(safe.address.replace("0x", "").rjust(64, "0"))
        signature_s = HexBytes(
            "41".rjust(64, "0")
        )  # Position of end of signature `0x41 == 65`
        signature_v = HexBytes("00")
        # First 32 bytes signature size, in this case 0
        contract_signature = HexBytes("0" * 64)
        signature = signature_r + signature_s + signature_v + contract_signature

        safe_signature = SafeSignature.parse_signature(signature, safe_tx_hash)[0]
        self.assertIsInstance(safe_signature, SafeSignatureContract)
        self.assertFalse(safe_signature.is_valid(self.ethereum_client, None))

        # Check with previously signedMessage
        tx = safe_contract.functions.signMessage(safe_tx_hash).build_transaction(
            {"from": safe.address}
        )
        safe_tx = safe.build_multisig_tx(safe.address, 0, tx["data"])
        safe_tx.sign(owner_1.key)
        safe_tx.execute(owner_1.key)

        safe_signature = SafeSignature.parse_signature(signature, safe_tx_hash)[0]
        self.assertTrue(safe_signature.is_valid(self.ethereum_client, None))
        self.assertIsInstance(safe_signature, SafeSignatureContract)

        # Check with crafted signature
        safe_tx_hash_2 = fast_keccak_text("test2")
        safe_signature = SafeSignature.parse_signature(signature, safe_tx_hash_2)[0]
        self.assertFalse(safe_signature.is_valid(self.ethereum_client, None))

        safe_tx_hash_2_message_hash = safe_contract.functions.getMessageHash(
            safe_tx_hash_2
        ).call()
        contract_signature = owner_1.signHash(safe_tx_hash_2_message_hash)["signature"]

        encoded_contract_signature_with_offset = encode_abi(
            ["bytes"], [contract_signature]
        )
        # {32 bytes - offset for the length}{32 bytes - length = 65 bytes}{65 bytes - content}
        self.assertEqual(len(encoded_contract_signature_with_offset), 160)
        # Safe dynamic part does not use the offset
        encoded_contract_signature = encoded_contract_signature_with_offset[32:]
        crafted_signature = (
            signature_r + signature_s + signature_v + encoded_contract_signature
        )
        safe_signature = SafeSignature.parse_signature(
            crafted_signature, safe_tx_hash_2
        )[0]
        self.assertEqual(contract_signature, safe_signature.contract_signature)
        self.assertTrue(safe_signature.is_valid(self.ethereum_client, None))

    def test_contract_multiple_signatures(self):
        """
        Test decode of multiple `CONTRACT_SIGNATURE` together
        """
        owner_1 = self.ethereum_test_account
        safe = self.deploy_test_safe_v1_1_1(
            owners=[owner_1.address], initial_funding_wei=Web3.to_wei(0.01, "ether")
        )
        safe_contract = safe.contract
        safe_tx_hash = fast_keccak_text("test")

        tx = safe_contract.functions.signMessage(safe_tx_hash).build_transaction(
            {"from": safe.address}
        )
        safe_tx = safe.build_multisig_tx(safe.address, 0, tx["data"])
        safe_tx.sign(owner_1.key)
        safe_tx.execute(owner_1.key)

        # Check multiple signatures. In this case we reuse signatures for the same owner, it won't make sense
        # in real life
        signature_r_1 = HexBytes(safe.address.replace("0x", "").rjust(64, "0"))
        signature_s_1 = HexBytes(
            "82".rjust(64, "0")
        )  # Position of end of signature `0x82 == (65 * 2)`
        signature_v_1 = HexBytes("00")
        contract_signature_1 = b""
        encoded_contract_signature_1 = encode_abi(["bytes"], [contract_signature_1])[
            32:
        ]  # It will {32 bytes offset}{32 bytes size}, we don't need offset

        signature_r_2 = HexBytes(safe.address.replace("0x", "").rjust(64, "0"))
        signature_s_2 = HexBytes(
            "a2".rjust(64, "0")
        )  # Position of end of signature `0xa2 == (65 * 2) + 32`
        signature_v_2 = HexBytes("00")
        safe_tx_hash_message_hash = safe_contract.functions.getMessageHash(
            safe_tx_hash
        ).call()
        contract_signature_2 = owner_1.signHash(safe_tx_hash_message_hash)["signature"]
        encoded_contract_signature_2 = encode_abi(["bytes"], [contract_signature_2])[
            32:
        ]  # It will {32 bytes offset}{32 bytes size}, we don't need offset

        signature = (
            signature_r_1
            + signature_s_1
            + signature_v_1
            + signature_r_2
            + signature_s_2
            + signature_v_2
            + encoded_contract_signature_1
            + encoded_contract_signature_2
        )

        count = 0
        for safe_signature, contract_signature in zip(
            SafeSignature.parse_signature(signature, safe_tx_hash),
            [contract_signature_1, contract_signature_2],
        ):
            self.assertEqual(safe_signature.contract_signature, contract_signature)
            self.assertTrue(safe_signature.is_valid(self.ethereum_client, None))
            self.assertEqual(
                safe_signature.signature_type, SafeSignatureType.CONTRACT_SIGNATURE
            )
            # Test exported signature
            exported_signature = SafeSignature.parse_signature(
                safe_signature.export_signature(), safe_tx_hash
            )[0]
            self.assertEqual(
                exported_signature.contract_signature, safe_signature.contract_signature
            )
            self.assertTrue(exported_signature.is_valid(self.ethereum_client, None))
            count += 1
        self.assertEqual(count, 2)



import logging
import os

from eth_account import Account
from eth_account.signers.local import LocalAccount
from web3 import Web3
from web3.contract import Contract
from web3.types import TxParams

from ..ethereum_client import EthereumClient, EthereumClientProvider
from ..multicall import Multicall
from .utils import deploy_erc20, send_tx

logger = logging.getLogger(__name__)


_cached_data = {
    "ethereum_client": None,  # Prevents initializing again
}


class EthereumTestCaseMixin:
    ethereum_client: EthereumClient
    ethereum_node_url: str
    w3: Web3
    ethereum_test_account: LocalAccount
    multicall: Multicall

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.ethereum_test_account = cls.get_ethereum_test_account()
        # Caching ethereum_client to prevent initializing again
        cls.ethereum_client = _cached_data["ethereum_client"]

        if not cls.ethereum_client:
            cls.ethereum_client = EthereumClientProvider()
            Multicall.deploy_contract(cls.ethereum_client, cls.ethereum_test_account)
            _cached_data["ethereum_client"] = cls.ethereum_client

        cls.ethereum_node_url = cls.ethereum_client.ethereum_node_url

        cls.w3 = cls.ethereum_client.w3
        cls.multicall = cls.ethereum_client.multicall
        assert cls.multicall, "Multicall must be defined"

    @classmethod
    def get_ethereum_test_account(cls):
        try:
            from django.conf import settings

            key = settings.ETHEREUM_TEST_PRIVATE_KEY
        except ModuleNotFoundError:
            key = os.environ.get(
                "ETHEREUM_TEST_PRIVATE_KEY",
                "b0057716d5917badaf911b193b12b910811c1497b5bada8d7711f758981c3773",  # Ganache account 9
            )
        return Account.from_key(key)

    @property
    def gas_price(self):
        return self.w3.eth.gas_price

    def send_tx(self, tx: TxParams, account: LocalAccount) -> bytes:
        return send_tx(self.w3, tx, account)

    def send_ether(self, to: str, value: int) -> bytes:
        return send_tx(self.w3, {"to": to, "value": value}, self.ethereum_test_account)

    def create_and_fund_account(
        self, initial_ether: float = 0, initial_wei: int = 0
    ) -> LocalAccount:
        account = Account.create()
        if initial_ether > 0.0 or initial_wei > 0:
            self.send_tx(
                {
                    "to": account.address,
                    "value": self.w3.to_wei(initial_ether, "ether") + initial_wei,
                },
                self.ethereum_test_account,
            )
        return account

    def deploy_erc20(
        self,
        name: str,
        symbol: str,
        owner: str,
        amount: int,
        decimals: int = 18,
    ) -> Contract:
        return deploy_erc20(
            self.w3,
            self.ethereum_test_account,
            name,
            symbol,
            owner,
            amount,
            decimals=decimals,
        )

    def deploy_example_erc20(self, amount: int, owner: str) -> Contract:
        return deploy_erc20(
            self.w3, self.ethereum_test_account, "Uxio", "UXI", owner, amount
        )



import os
from functools import lru_cache
from typing import Union

import eth_abi
from eth._utils.address import generate_contract_address
from eth_account import Account
from eth_typing import Address, AnyAddress, ChecksumAddress, Hash32, HexAddress, HexStr
from eth_utils import to_normalized_address
from hexbytes import HexBytes
from sha3 import keccak_256
from web3.types import TxParams, Wei


def get_empty_tx_params() -> TxParams:
    """
    :return: Empty tx params, so calls like `build_transaction` don't call the RPC trying to get information
    """
    return {
        "gas": Wei(1),
        "gasPrice": Wei(1),
    }


@lru_cache(maxsize=int(os.getenv("CACHE_KECCAK", 512)))
def _keccak_256(value: bytes) -> bytes:
    return keccak_256(value)


def fast_keccak(value: bytes) -> Hash32:
    """
    Calculates ethereum keccak256 using fast library `pysha3`

    :param value:
    :return: Keccak256 used by ethereum as `HexBytes`
    """
    return HexBytes(_keccak_256(value).digest())


def fast_keccak_text(value: str) -> Hash32:
    """
    Calculates ethereum keccak256 using fast library `pysha3`

    :param value:
    :return: Keccak256 used by ethereum as `HexBytes`
    """
    return fast_keccak(value.encode())


def fast_keccak_hex(value: bytes) -> HexStr:
    """
    Same as `fast_keccak`, but it's a little more optimal calling `hexdigest()`
    than calling `digest()` and then `hex()`

    :param value:
    :return: Keccak256 used by ethereum as a hex string (not 0x prefixed)
    """
    return HexStr(_keccak_256(value).hexdigest())


def _build_checksum_address(
    norm_address: HexStr, address_hash: HexStr
) -> ChecksumAddress:
    """
    https://github.com/ethereum/EIPs/blob/master/EIPS/eip-55.md

    :param norm_address: address in lowercase (not 0x prefixed)
    :param address_hash: keccak256 of `norm_address` (not 0x prefixed)
    :return:
    """
    return ChecksumAddress(
        HexAddress(
            HexStr(
                "0x"
                + (
                    "".join(
                        (
                            norm_address[i].upper()
                            if int(address_hash[i], 16) > 7
                            else norm_address[i]
                        )
                        for i in range(0, 40)
                    )
                )
            )
        )
    )


@lru_cache(maxsize=int(os.getenv("CACHE_CHECKSUM_ADDRESS", 1_000_000_000)))
def _fast_to_checksum_address(address: HexAddress):
    address_hash = fast_keccak_hex(address.encode())
    return _build_checksum_address(address, address_hash)


def fast_to_checksum_address(value: Union[AnyAddress, str, bytes]) -> ChecksumAddress:
    """
    Converts to checksum_address. Uses more optimal `pysha3` instead of `eth_utils` for keccak256 calculation

    :param value:
    :return:
    """
    if isinstance(value, bytes):
        if len(value) != 20:
            raise ValueError(
                "Cannot convert %s to a checksum address, 20 bytes were expected"
            )

    norm_address = HexAddress(HexStr(to_normalized_address(value)[2:]))
    return _fast_to_checksum_address(norm_address)


def fast_bytes_to_checksum_address(value: bytes) -> ChecksumAddress:
    """
    Converts to checksum_address. Uses more optimal `pysha3` instead of `eth_utils` for keccak256 calculation.
    As input is already in bytes, some checks and conversions can be skipped, providing a speedup of ~50%

    :param value:
    :return:
    """
    if len(value) != 20:
        raise ValueError(
            "Cannot convert %s to a checksum address, 20 bytes were expected"
        )
    norm_address = HexAddress(HexStr(bytes(value).hex()))
    return _fast_to_checksum_address(norm_address)


def fast_is_checksum_address(value: Union[AnyAddress, str, bytes]) -> bool:
    """
    Fast version to check if an address is a checksum_address

    :param value:
    :return: `True` if checksummed, `False` otherwise
    """
    if not isinstance(value, str) or len(value) != 42 or not value.startswith("0x"):
        return False
    try:
        return fast_to_checksum_address(value) == value
    except ValueError:
        return False


def get_eth_address_with_invalid_checksum() -> str:
    address = Account.create().address
    return "0x" + "".join(
        [c.lower() if c.isupper() else c.upper() for c in address[2:]]
    )


def decode_string_or_bytes32(data: bytes) -> str:
    try:
        return eth_abi.decode(["string"], data)[0]
    except (OverflowError, eth_abi.exceptions.DecodingError):
        name = eth_abi.decode(["bytes32"], data)[0]
        end_position = name.find(b"\x00")
        if end_position == -1:
            return name.decode()
        else:
            return name[:end_position].decode()


def remove_swarm_metadata(code: bytes) -> bytes:
    """
    Remove swarm metadata from Solidity bytecode

    :param code:
    :return: Code without metadata
    """
    swarm = b"\xa1\x65bzzr0"
    position = code.rfind(swarm)
    if position == -1:
        raise ValueError("Swarm metadata not found in code %s" % code.hex())
    return code[:position]


def compare_byte_code(code_1: bytes, code_2: bytes) -> bool:
    """
    Compare code, removing swarm metadata if necessary

    :param code_1:
    :param code_2:
    :return: True if same code, False otherwise
    """
    if code_1 == code_2:
        return True
    else:
        codes = []
        for code in (code_1, code_2):
            try:
                codes.append(remove_swarm_metadata(code))
            except ValueError:
                codes.append(code)

        return codes[0] == codes[1]


def mk_contract_address(address: Union[str, bytes], nonce: int) -> ChecksumAddress:
    """
    Generate expected contract address when using EVM CREATE

    :param address:
    :param nonce:
    :return:
    """
    return fast_to_checksum_address(
        generate_contract_address(Address(HexBytes(address)), nonce)
    )


def mk_contract_address_2(
    from_: Union[ChecksumAddress, bytes],
    salt: Union[HexStr, bytes],
    init_code: Union[HexStr, bytes],
) -> ChecksumAddress:
    """
    Generate expected contract address when using EVM CREATE2.

    :param from_: The address which is creating this new address (need to be 20 bytes)
    :param salt: A salt (32 bytes)
    :param init_code: A init code of the contract being created
    :return: Address of the new contract
    """

    from_ = HexBytes(from_)
    salt = HexBytes(salt)
    init_code = HexBytes(init_code)

    assert len(from_) == 20, f"Address {from_.hex()} is not valid. Must be 20 bytes"
    assert len(salt) == 32, f"Salt {salt.hex()} is not valid. Must be 32 bytes"
    assert len(init_code) > 0, f"Init code {init_code.hex()} is not valid"

    init_code_hash = fast_keccak(init_code)
    contract_address = fast_keccak(HexBytes("ff") + from_ + salt + init_code_hash)
    return fast_bytes_to_checksum_address(contract_address[12:])
